/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <dlfcn.h>
#include <signal.h>

#include <string>
#include <fstream>
#include <vector>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "utils.h"
#include "log.h"
#include "launcher.h"
#include "dotnet_launcher.h"

#define PLUGIN_PATH "/usr/share/dotnet.tizen/lib/libdotnet_plugin.so"

namespace tizen {
namespace runtime {
namespace dotnetcore {

#if defined (__aarch64__)
#define ARCHITECTURE_IDENTIFIER "arm64"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-arm64", "linux", "unix-arm64", "unix", "any", "base"};

#elif defined (__arm__)
#define ARCHITECTURE_IDENTIFIER "arm"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"tizen.4.0.0-armel", "tizen.4.0.0", "tizen-armel", "tizen", "linux-armel", "linux", "unix-armel", "unix", "any", "base"};

#elif defined (__x86_64__)
#define ARCHITECTURE_IDENTIFIER "x64"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-x64", "linux", "unix-x64", "unix", "any", "base"};

#elif defined (__i386__)
#define ARCHITECTURE_IDENTIFIER "x86"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-x86", "linux", "unix-x86", "unix", "any", "base"};

#else
#error "Unknown target"
#endif

static std::string getExtraNativeLibDirs(const std::string& appRoot)
{
	std::string candidate;
	for (unsigned int i = 0; i < RID_FALLBACK_GRAPH.size(); i++) {
		if(!candidate.empty()) {
			candidate += ":";
		}
		candidate += concatPath(appRoot, "bin/runtimes/" + RID_FALLBACK_GRAPH[i] + "/native");
	}

	candidate = candidate + ":" + concatPath(appRoot, "lib/" ARCHITECTURE_IDENTIFIER);
	if (!strncmp(ARCHITECTURE_IDENTIFIER, "arm64", 5)) {
		candidate = candidate + ":" + concatPath(appRoot, "lib/aarch64");
	}

	return candidate;
}

typedef void (*mm_dotnet_cleanup_fn)(int signum);
static void cleanupMM(int signum)
{
	void* handle = dlopen("/usr/lib/libmmfsession.so.0.0.0", RTLD_NOW);
	if (handle != NULL) {
		mm_dotnet_cleanup_fn cleanup = (mm_dotnet_cleanup_fn)dlsym(handle, "mm_session_dotnet_cleanup");
		if (cleanup) {
			cleanup(signum);
		}

		dlclose(handle);
	}

	handle = dlopen("/usr/lib/libmmfsound.so.0.1.0", RTLD_NOW);
	if (handle != NULL) {
		mm_dotnet_cleanup_fn cleanup = (mm_dotnet_cleanup_fn)dlsym(handle, "mm_sound_dotnet_cleanup");
		if (cleanup) {
			cleanup(signum);
		}

		dlclose(handle);
	}
}

struct sigaction sig_abrt_new;
struct sigaction sig_abrt_old;

static bool checkOnSigabrt = false;
static void onSigabrt(int signum)
{
	// ignore return value of write().
	// this function is called when process go to die
	// there is no need to handle return value for logging.
	write(2, "onSigabrt called\n", 17);

	if (checkOnSigabrt) {
		write(2, "onSigabrt called again. go to exit\n", 35);
		exit(0);
	}

	cleanupMM(signum);

	if (hasException()) {
		write(2, "******************************************************\n", 55);
		write(2, "Unhandled exception is occured. check application code\n", 55);
		write(2, "******************************************************\n", 55);
	}

	checkOnSigabrt = true;
	if (sigaction(SIGABRT, &sig_abrt_old, NULL) == 0) {
		if (raise(signum) < 0) {
			write(2, "Fail to raise SIGABRT\n", 22);
		}
	} else {
		write(2, "Fail to set original SIGABRT handler\n", 37);
	}
}

static void registerSigHandler()
{
	sig_abrt_new.sa_handler = onSigabrt;
	if (sigemptyset(&sig_abrt_new.sa_mask) != 0) {
		_ERR("Fail to sigemptyset");
	}

	if (sigaction(SIGABRT, &sig_abrt_new, &sig_abrt_old) < 0) {
		_ERR("Fail to add sig handler");
	}
}

CoreRuntime::CoreRuntime() :
	initializeClr(nullptr),
	executeAssembly(nullptr),
	shutdown(nullptr),
	createDelegate(nullptr),
	__coreclrLib(nullptr),
	__hostHandle(nullptr),
	__domainId(-1),
	preparedFunction(nullptr),
	launchFunction(nullptr),
	__pluginLib(nullptr),
	pluginInitialize(nullptr),
	pluginPreload(nullptr),
	pluginSetAppInfo(nullptr),
	pluginSetCoreclrInfo(nullptr),
	pluginGetDllPath(nullptr),
	pluginBeforeExecute(nullptr),
	pluginFinalize(nullptr),
	fd(0)
{
	if (runLoggingThread() < 0) {
		_ERR("Failed to create logging thread");
	}

#define __XSTR(x) #x
#define __STR(x) __XSTR(x)

#ifdef DEVICE_API_DIR
	__deviceAPIDirectory = __STR(DEVICE_API_DIR);
#endif
#ifdef RUNTIME_DIR
	__runtimeDirectory = __STR(RUNTIME_DIR);
#endif
#ifdef NATIVE_LIB_DIR
	__nativeLibDirectory = __STR(NATIVE_LIB_DIR);
#endif

#undef __STR
#undef __XSTR

	// support launcher plugin
	if (!fileNotExist(PLUGIN_PATH)) {
		__pluginLib = dlopen(PLUGIN_PATH, RTLD_NOW | RTLD_LOCAL);
		if (__pluginLib) {
			pluginInitialize  = (plugin_initialize_ptr)dlsym(__pluginLib, "plugin_initialize");
			pluginPreload  = (plugin_preload_ptr)dlsym(__pluginLib, "plugin_preload");
			pluginSetAppInfo  = (plugin_set_app_info_ptr)dlsym(__pluginLib, "plugin_set_app_info");
			pluginSetCoreclrInfo = (plugin_set_coreclr_info_ptr)dlsym(__pluginLib, "plugin_set_coreclr_info");
			pluginGetDllPath  = (plugin_get_dll_path_ptr)dlsym(__pluginLib, "plugin_get_dll_path");
			pluginBeforeExecute  = (plugin_before_execute_ptr)dlsym(__pluginLib, "plugin_before_execute");
			pluginFinalize  = (plugin_finalize_ptr)dlsym(__pluginLib, "plugin_finalize");
		}
	}

	if (pluginInitialize)
		pluginInitialize();

	_DBG("Constructor called!!");
}

CoreRuntime::~CoreRuntime()
{
	dispose();
}

int CoreRuntime::initialize(bool standalone)
{
#ifdef __arm__
	// libunwind library is used to unwind stack frame, but libunwind for ARM
	// does not support ARM vfpv3/NEON registers in DWARF format correctly.
	// Therefore let's disable stack unwinding using DWARF information
	// See https://github.com/dotnet/coreclr/issues/6698
	//
	// libunwind use following methods to unwind stack frame.
	// UNW_ARM_METHOD_ALL          0xFF
	// UNW_ARM_METHOD_DWARF        0x01
	// UNW_ARM_METHOD_FRAME        0x02
	// UNW_ARM_METHOD_EXIDX        0x04
	putenv(const_cast<char *>("UNW_ARM_UNWIND_METHOD=6"));
#endif // __arm__

	// mm-session / mm-sound check application type with environment setting.
	putenv(const_cast<char *>("DOTNET_APP=1"));

	if (__deviceAPIDirectory.empty()) {
		_ERR("Empty Device API Directory");
		return 1;
	} else {
		__deviceAPIDirectory = absolutePath(__deviceAPIDirectory);
	}

	if (__runtimeDirectory.empty()) {
		_ERR("Empty Runtime Directory");
		return 1;
	} else {
		__runtimeDirectory = absolutePath(__runtimeDirectory);
	}

	// set Reference API directory
	__refAPIDirectory = __deviceAPIDirectory + "/ref";

	std::string libCoreclr(concatPath(__runtimeDirectory, "libcoreclr.so"));

	_DBG("libcoreclr : %s", libCoreclr.c_str());

	__coreclrLib = dlopen(libCoreclr.c_str(), RTLD_NOW | RTLD_LOCAL);
	if (__coreclrLib == nullptr) {
		char *err = dlerror();
		_ERR("dlopen failed to open libcoreclr.so with error %s", err);
		return 1;
	}

#define CORELIB_RETURN_IF_NOSYM(type, variable, name) \
	do { \
		variable = (type)dlsym(__coreclrLib, name); \
		if (variable == nullptr) { \
			_ERR(name " is not found in the libcoreclr.so"); \
			return 1; \
		} \
	} while (0)

	CORELIB_RETURN_IF_NOSYM(coreclr_initialize_ptr, initializeClr, "coreclr_initialize");
	CORELIB_RETURN_IF_NOSYM(coreclr_execute_assembly_ptr, executeAssembly, "coreclr_execute_assembly");
	CORELIB_RETURN_IF_NOSYM(coreclr_shutdown_ptr, shutdown, "coreclr_shutdown");
	CORELIB_RETURN_IF_NOSYM(coreclr_create_delegate_ptr, createDelegate, "coreclr_create_delegate");

#undef CORELIB_RETURN_IF_NOSYM

	_DBG("libcoreclr dlopen and dlsym success");

	if (!standalone && pluginPreload)
		pluginPreload();

	fd = open("/proc/self", O_DIRECTORY);
	std::string path_tmp = std::string("/proc/self/fd/") + std::to_string(fd);

	std::string appRoot = path_tmp;
	std::string appBin = concatPath(appRoot, "bin");
	std::string appLib = concatPath(appRoot, "lib");
	std::string probePath = appBin + ":" + appLib;

	std::string tpa;
	std::vector<std::string> searchDirectories;
	searchDirectories.push_back(__runtimeDirectory);
	searchDirectories.push_back(__deviceAPIDirectory);
	searchDirectories.push_back(__refAPIDirectory);

	if (pluginGetDllPath) {
		std::string pluginPath = pluginGetDllPath();
		if (!pluginPath.empty()) {
			searchDirectories.push_back(pluginPath);
		}
	}

	assembliesInDirectory(searchDirectories, tpa);

	std::string nativeLibPath;
	nativeLibPath = getExtraNativeLibDirs(appRoot) + ":" + appBin + ":" + appLib + ":" + __nativeLibDirectory;

	std::string appName = std::string("dotnet-launcher-") + std::to_string(getpid());
	if (!initializeCoreClr(appName.c_str(), probePath.c_str(), nativeLibPath.c_str(), tpa.c_str())) {
		_ERR("Failed to initialize coreclr");
		return 1;
	}

	return 0;
}

bool CoreRuntime::initializeCoreClr(const char* appId,
									 const char* assemblyProbePaths,
									 const char* pinvokeProbePaths,
									 const char* tpaList)
{
	const char *propertyKeys[] = {
		"TRUSTED_PLATFORM_ASSEMBLIES",
		"APP_PATHS",
		"APP_NI_PATHS",
		"NATIVE_DLL_SEARCH_DIRECTORIES",
		"AppDomainCompatSwitch"
	};

	const char *propertyValues[] = {
		tpaList,
		assemblyProbePaths,
		assemblyProbePaths,
		pinvokeProbePaths,
		"UseLatestBehaviorWhenTFMNotSpecified"
	};

	std::string selfPath = readSelfPath();

	int st = initializeClr(selfPath.c_str(),
							appId,
							sizeof(propertyKeys) / sizeof(propertyKeys[0]),
							propertyKeys,
							propertyValues,
							&__hostHandle,
							&__domainId);

	if (st < 0) {
		_ERR("initialize core clr fail! (0x%08x)", st);
		return false;
	}

	if (pluginSetCoreclrInfo)
		pluginSetCoreclrInfo(__hostHandle, __domainId);

	_DBG("Initialize core clr success");
	return true;
}

void CoreRuntime::dispose()
{
	if (__hostHandle != nullptr) {
		int st = shutdown(__hostHandle, __domainId);
		if (st < 0)
			_ERR("shutdown core clr fail! (0x%08x)", st);
	}

	if (dlclose(__coreclrLib) != 0)
		_ERR("libcoreclr.so close failed");

	__coreclrLib = nullptr;

	if (pluginFinalize)
		pluginFinalize();

	if (__pluginLib != nullptr) {
		if (dlclose(__pluginLib) != 0)
			_ERR("libdotnet_plugin.so close failed");

		__pluginLib = nullptr;
		pluginInitialize = nullptr;
		pluginPreload = nullptr;
		pluginSetAppInfo = nullptr;
		pluginSetCoreclrInfo = nullptr;
		pluginGetDllPath = nullptr;
		pluginBeforeExecute = nullptr;
		pluginFinalize = nullptr;
	}

	_DBG("Dotnet runtime disposed");
}

int CoreRuntime::launch(const char* appId, const char* root, const char* path, int argc, char* argv[])
{
	if (path == nullptr) {
		_ERR("executable path is null");
		return 1;
	}

	if (fileNotExist(path)) {
		_ERR("File not exist : %s", path);
		return 1;
	}

	if (redirectFD() < 0) {
		_ERR("Fail to redirect FD");
                return 1;
	}

	registerSigHandler();

	if (pluginSetAppInfo)
		pluginSetAppInfo(appId, path);

	int fd2 = open(root, O_DIRECTORY);
	dup3(fd2, fd, O_CLOEXEC);
	if (fd2 >= 0)
		close(fd2);

	if (pluginBeforeExecute)
		pluginBeforeExecute();

	unsigned int ret = 0;
	int st = executeAssembly(__hostHandle, __domainId, argc, (const char**)argv, path, &ret);
	if (st < 0)
		_ERR("Failed to Execute Assembly %s (0x%08x)", path, st);
	return ret;
}

}  // namespace dotnetcore
}  // namespace runtime
}  // namespace tizen
